@@ -1,16 +1,28 @@
 /****************************************************************************
  * util.c
  * Conversion routines between S-Lang and Perl data types.
+ *
+ * needs better organisation/planning!
+ * - should we have a table of types and functions that
+ *   support them, since this would make changes easier/code
+ *   more modular, at a possible speed expense?
+ *
+ * Need to change to return an "opaque" reference for un-defined
+ * types (sub-types of Inline::SLang::type say?) so that we can
+ * pass around items we don't know about (such as Region_Type
+ * from CIAO's region library)
+ *
  ****************************************************************************/
 
 /* Supported types (taken from slang.h):
 
 2p - does conversion from S-Lang to Perl work?
+2s - does conversion from Perl to S-Lang work?
+
      s - scalars
      1 - only 1D arrays
      2 - <= 2D arrays
      n - nD arrays
-2s - does conversion from Perl to S-Lang work?
 
 
 2p 2s S-Lang                   Perl
@@ -39,13 +51,14 @@ s2 s  SLANG_COMPLEX_TYPE       Math::Complex
 
 ?     SLANG_ASSOC_TYPE         hash reference
 
-s  s  SLANG_DATATYPE_TYPE      Inline::SLang::datatype
+s     SLANG_STRUCT_TYPE        Inline::SLang::struct
+
+s2 s  SLANG_DATATYPE_TYPE      Inline::SLang::datatype
 
 ?     SLANG_FILE_PTR_TYPE      should be able to convert to a perl IO variable
 
- */
-
-/* Are we going to support these?
+Are we going to support these?
+------------------------------
 
 SLANG_ANY_TYPE - should be easy to support since can just convert
   to the correct perl type BUT would we want to keep the
@@ -57,18 +70,13 @@ SLANG_UNDEFINED_TYPE - convert to undef ?
 SLANG_VOID_TYPE
 SLANG_REF_TYPE
 
-SLANG_STRUCT_TYPE - handle as an object?
-SLANG_ISTRUCT_TYPE - what is this - an intrinsic structure>
-
+SLANG_ISTRUCT_TYPE - what is this - an intrinsic structure?
 
 SLANG_BSTRING_TYPE - what is this?
 SLANG_FILE_FD_TYPE - what is this?
 
- */
-
-/*
- * Unsupported types (taken from slang.h):
-
+Unsupported types (taken from slang.h):
+---------------------------------------
 SLANG_INTP_TYPE - as apparently I shouldn't see one of these
 
  *
@@ -76,50 +84,33 @@ SLANG_INTP_TYPE - as apparently I shouldn't see one of these
 
 #include "util.h"
 
+/* used by the CALL_xxx macros */
+static char *
+_get_object_type( SV *obj ) {
+  HV *stash = SvSTASH( obj ); /* assume obj really is an object */
+  return ( stash ? HvNAME(stash) : "<none>" );
+} /* _get_object_type() */
+
 /*
- * utility routines for pl2sl()
+ * since we use $1..$n to create/manipulate variables
+ * it can lead to mis-leading results (ie what you think a routine
+ * is doing is only happening because $1 happens to be set
+ * correctly from a previous call [can happen if test set/get
+ * routines next to each other].
  *
- * from_perl_complex() - convert Math::Complex object to Real,Imag numbers
- *
- * the array  should handle the integers as does the scalar
- * code (ie respect the types)
+ * so, we try and clean things out (setting them to null)
  */
+static void
+_clean_slang_vars( int n ) {
+  char stxt[12]; /* assuming n <= 99 */
+  int i;
+  for ( i = 1; i <= n; i++ ) {
+    (void) sprintf( stxt, "$%d = NULL;", i );
+    if ( -1 == SLang_load_string( stxt ) )
+      croak( "Internal error: unable to clean up S-Lang $n vars\n" );
+  }
+} /* _clean_slang_vars() */
 
-static double
-call_double_complex_method( SV *item, const char *method ) {
-
-  dSP; /* this defines variables; it's not a bit of code! */
-  double result;
-  int count;
-
-  /*
-   * - see 'perldoc perlcall' for information on what's happening
-   */
-
-  ENTER;
-  SAVETMPS;
-  PUSHMARK(SP);
-
-  /* tried caching "Math::Complex", but it got over-written somewhere */
-  XPUSHs( item );
-  PUTBACK;
-
-  Printf( ("Calling Math::Complex->%s()\n",method) );
-  count = call_method( method, G_SCALAR );
-  
-  SPAGAIN;
-  if ( count != 1 )
-    croak("Math::Complex method did not return a value (unable to convert Complex_Type)\n");
-
-  result = (double) POPn;
-
-  PUTBACK;
-  FREETMPS;
-  LEAVE;
-
-  return result;
-
-} /* call_double_complex_method() */
 
 /*
  * convert perl variables to S-Lang variables
@@ -128,6 +119,11 @@ call_double_complex_method( SV *item, const char *method ) {
  * - this will probably turn out to be a bad idea; for instance it
  *   means it can't be called recursively when converting
  *   array/associative arrays.
+ *
+ * - push NULL onto the stack for those variables we do not understand
+ *   also - as an aid to debugging - print messages to STDERR
+ *   NEED to work out how to inform the calling program that this
+ *   transformation has taken place
  */
 
 void
@@ -139,7 +135,7 @@ pl2sl( SV *item ) {
     if ( -1 == SLang_push_null() )
       croak( "Unable to push a NULL onto the stack" );
     return;
-  }
+  } /* undef */
 
   /* integer */
   if ( SvIOK(item) ) {
@@ -147,17 +143,17 @@ pl2sl( SV *item ) {
     if ( -1 == SLang_push_integer( SvIV(item) ) )
       croak( "Unable to push an integer onto the stack" );
     return;
-  }
+  } /* integer */
 
-  /* floating-point number */
+  /* floating-point */
   if ( SvNOK(item) ) {
     Printf( ("item=float %f\n", SvNV(item)) );
     if ( -1 == SLang_push_double( SvNV(item) ) )
       croak( "Unable to push a double onto the stack" );
     return;
-  }
+  } /* floating-point */
 
-  /* a string */
+  /* string */
   if ( SvPOK(item) ) {
     STRLEN len;
     char *ptr = SvPV(item, len);
@@ -165,25 +161,25 @@ pl2sl( SV *item ) {
     if ( -1 == SLang_push_string( ptr ) )
       croak( "Unable to push a string onto the stack" );
     return;
-  } /* string 2 S-Lang */
+  } /* string */
 
-  /* a complex number (Math::Complex object) */
+  /* Math::Complex */
   if ( sv_isobject(item) && sv_derived_from(item, "Math::Complex" ) ) {
     double real, imag;
 
     Printf( ("*** converting Perl's Math::Complex to S-Lang Complex_Type\n") );
 
     /* call the Re and Im methods */
-    real = call_double_complex_method( item, "Re" );
-    imag = call_double_complex_method( item, "Im" );
+    CALL_METHOD_SCALAR_DOUBLE( item, "Re", , real );
+    CALL_METHOD_SCALAR_DOUBLE( item, "Im", , imag );
 
     /* push the complex number onto the S-Lang stack */
     if ( -1 == SLang_push_complex( real, imag ) )
       croak( "Unable to push a complex number onto the stack" );
     return;
-  } /* complex 2 S-Lang */
+  } /* Math::Complex */
 
-  /* an Inline::SLang::datatype object */
+  /* Inline::SLang::datatype */
   if ( sv_isobject(item) && sv_derived_from(item, "Inline::SLang::datatype" ) ) {
     char *name;
 
@@ -197,13 +193,93 @@ pl2sl( SV *item ) {
      * convert to the S-Lang datatype. All we do is push the name
      * onto the stack and let S-Lang do the conversion to an
      * actual DataType_Type
+     * - not the most efficient implementation but saves messing
+     *   with the internals of S-Lang
      */
-    if ( -1 == SLang_load_string( name ) ) /* ignore the trailing ';' that we should have */
+    if ( -1 == SLang_load_string( name ) ) /* ignores the trailing ';' that we should have */
       croak( "Unable to create a DataType_Type on the stack" );
     return;
-  } /* datatype 2 S-Lang */
+  } /* Inline::SLang::datatype */
 
-  /* an array: always set to SLANG_ANY_TYPE */
+  /*
+   * Inline::SLang::struct
+   */
+  if ( sv_isobject(item) && sv_derived_from(item, "Inline::SLang::struct" ) ) {
+
+    SV *dstruct;
+    SV *aref;
+    AV *fields;
+    I32 nfields, i;
+
+    Printf( ("*** converting Perl's Inline::SLang::struct to S-Lang Struct_Type\n") );
+    fixme( "memleak" );
+
+    /*
+     * should we use the API of the perl object or just access the hash array
+     * directly?
+     *
+     * we actually use a bit of both - not the cleanest/most ideal way of
+     * doing this, but it's a bit painful as we want to convert the
+     * fields from their perl representation to their S-Lang one
+     *
+     * one possible improveement (if it works) is to only call
+     * Inline::SLang::struct's get() method once - ie stick all
+     * the items onto the Perl stack in one go
+     */
+
+    /* create a structure in $1 with the correct fields */
+    CALL_METHOD_SCALAR_SV( item, "_define_struct", , dstruct );
+    Printf( ("struct definition =\n[%s]\n", SvPV_nolen(dstruct)) );
+    if ( -1 == SLang_load_string( SvPV_nolen(dstruct) ) )
+      croak("Internal Error: unable to create a struct in $1\n");
+
+    /* return the field names (as a string array) */
+    CALL_METHOD_SCALAR_SV( item, "fields", , aref );
+    //    aref = sv_2mortal( aref ); // randomly trying things
+    fields = (AV *) SvRV( aref );
+    nfields = 1 + av_len( fields );
+
+    /* MEM:
+       have tried 'aref = sv_2mortal( aref );'
+       and 'SvREFCNT_inc( *name );'
+       - umm, need to look at what call_method_scalar_sv returns
+         (is there an increased ref count?)
+    */
+
+    /* loop through each field and set its value */
+    for ( i = 0; i < nfields; i++ ) {
+      SV **name;
+      SV *value;
+
+      name = av_fetch( fields, i, 0 );
+      // SvREFCNT_inc( *name ); // randomly trying things
+      if ( -1 == SLang_push_string( SvPV_nolen(*name) ) )
+	croak( "Unable to push a string onto the stack" );
+
+      CALL_METHOD_SCALAR_SV( item, "get", C2PL_MARG( *name );, value );
+      // pl2sl( sv_2mortal(value) );  // randomly trying things
+      pl2sl( value );
+      
+      if ( -1 ==
+	   SLang_load_string( "$2=(); $3=(); set_struct_field( $1, $3, $2 );" ) )
+	croak( "Internal Error:\n"
+               " Unable to fill in a field (in $3) of a structure (in $1)"
+               " with the value in $2\n" );
+    }
+    
+    /* 
+     * need to pop item off S-Lang's internal stack and push
+     * it onto S-Lang's main stack (or I've messed up above)
+     */
+    if ( -1 == SLang_load_string("$2=struct {value};set_struct_field($2,\"value\",$1);__push_args($2);") )
+      croak("Error: unable to hack a structure together\n");
+
+    _clean_slang_vars(3);
+    return;
+
+  } /* Inline::SLang::struct */
+
+  /* an array reference: always set to SLANG_ANY_TYPE */
   if ( SvROK(item) && SvTYPE(SvRV(item)) == SVt_PVAV ) {
     AV *av = (AV*) SvRV(item);
     SV *elem;
@@ -244,16 +320,13 @@ pl2sl( SV *item ) {
     ***/
 
     return;
-  }
+  } /* array reference */
 
   /*
    * hash ref
    *
-   * it doesn't appear that there's much of an API to access S-Lang's
-   * associative arrays from C. We could cheat and do it in S-Lang,
-   * but that means creating a temporary variable (or having one
-   * that we can use [with down-stream issues regarding threading])
-   *
+   * there isn't an API to access S-Lang's associative arrays from C. 
+   * So it looks like we'll have to let S-Lang do it for us.
    */
   if ( SvROK(item) && SvTYPE(SvRV(item)) == SVt_PVHV ) {
     HV *hv = (HV*) SvRV(item);
@@ -264,6 +337,8 @@ pl2sl( SV *item ) {
 
     croak( "Internal Error: currently unable to convert perl hash references\n to S-Lang associative arrays" );
 
+    /* create an assoc type array containint Any_Type variables */
+
     /***
 	o = rb_hash_new();
 
@@ -277,28 +352,34 @@ pl2sl( SV *item ) {
 	}
     ***/
 
-  }
+  } /* hash reference */
 
-  /*
-   * should this be before all the checks so we can catch
-   * 'perl scalar blessed into an object' situations?
-   *
-   * and shouldn't we have check it's a reference first?
-   */
-  if ( sv_isobject(item) ) {
-    croak( "Sorry, I am unable to convert a perl object into a S-Lang datatype\n" );
-  }
+    /*
+     * if we've got this far then we return NULL
+     * . we could croak()
+     * . we could call an exception
+     * . we could set an error flag
+     *
+     * ACTUALLY, I think we should probably return a reference to the
+     * type - perhaps Inline::SLang::unknown, say,
+     * so that we can pass around - if not handle - unknown types
+     *
+     * the screen messages are temporary
+     */
+    if ( sv_isobject(item) ) {
+      fprintf( stderr, "Internal error: unable to handle objects of type %s\n",
+         _get_object_type(item) );
+    } else
+      fprintf( stderr, "Internal Error: unable to handle this type of perl variable\n" );
 
-  croak( "Internal Error: trying to convert a 'strange' perl type to S-Lang\n" );
+    if ( -1 == SLang_push_null() )
+      croak( "Unable to push a NULL onto the stack" );
 
 } /* pl2sl() */
 
 /*
  * utility routines for sl2pl()
  *
- * make_perl_complex() - convert Real,Imag to a Math::Complex object
- * from_perl_complex() - convert Math::Complex object to Real,Imag numbers
- *
  * sl2pl1D() - convert a 1D array into a perl array reference
  * sl2pl2D() - convert a 2D array into a perl array reference
  * sl2plND() - convert a nD array into a perl array reference
@@ -307,101 +388,6 @@ pl2sl( SV *item ) {
  * code (ie respect the types)
  */
 
-static SV *
-make_perl_complex( double real, double imag ) {
-
-  dSP; /* this defines variables; it's not a bit of code! */
-  SV *object;
-  int count;
-
-  /*
-   * create the Math::Complex object
-   * - see 'perldoc perlcall' for information on what's happening
-   */
-
-  ENTER;
-  SAVETMPS;
-  PUSHMARK(SP);
-
-  /* tried caching "Math::Complex", but it got over-written somewhere */
-  XPUSHs(sv_2mortal(newSVpv("Math::Complex",0)));
-  XPUSHs(sv_2mortal(newSVnv(real)));
-  XPUSHs(sv_2mortal(newSVnv(imag)));
-  PUTBACK;
-
-  Printf( ("Calling Math::Complex->make()\n") );
-  count = call_method( "make", G_SCALAR );
-  
-  SPAGAIN;
-  if ( count != 1 )
-    croak("Math::Complex did not return an object (unable to convert Complex_Type)\n") ;
-
-  /*
-   * this is unlikely to be the right thing to do, but we need
-   * to do something to stop the object from disappearing
-   * once the stack is restored
-   */
-  object = SvREFCNT_inc( POPs );
-
-  PUTBACK;
-  FREETMPS;
-  LEAVE;
-
-  return object;
-
-} /* make_perl_complex */
-
-/*
- * SL2PL_ARRAY1D_ITYPE( INT, int ) 
- * will create code to handle 1D array values of integer types
- *
- *   SLANG_INT_TYPE and SLANG_UINT_TYPE
- *
- * SL2PL_ARRAY1D_FTYPE( FLOAT, float ) 
- * will create code to handle 1D array values of float types
- *
- *   SLANG_FLOAT_TYPE 
- *
- * it assumes that i is defined as an int
- * and the array is in at and that the perl array
- * is in parray
- */
-
-#define SL2PL_ARRAY1D_ITYPE(stypeu,ctype) \
-  case SLANG_##stypeu##_TYPE: \
-    { \
-      ctype ival; \
-      for ( i = 0; i < at->num_elements; i++ ) { \
-	(void) SLang_get_array_element( at, &i, &ival ); \
-	Printf( ("  ctype array1D[%d] = %d\n", i, ival ) ); \
-	av_store( parray, i, newSViv( ival ) ); \
-      } \
-      break; \
-    } \
-  \
-  case SLANG_U##stypeu##_TYPE: \
-    { \
-      unsigned ctype ival; \
-      for ( i = 0; i < at->num_elements; i++ ) { \
-	(void) SLang_get_array_element( at, &i, &ival ); \
-	Printf( ("  unsigned ctype array1D[%d] = %d\n", i, ival ) ); \
-	av_store( parray, i, newSVuv( ival ) ); \
-      } \
-      break; \
-    }
-
-#define SL2PL_ARRAY1D_FTYPE(stypeu,ctype) \
-  case SLANG_##stypeu##_TYPE: \
-    { \
-      ctype fval; \
-      for ( i = 0; i < at->num_elements; i++ ) { \
-	(void) SLang_get_array_element( at, &i, &fval ); \
-	Printf( ("  ctype array1D[%d] = %g\n", i, fval ) ); \
-	av_store( parray, i, newSVnv( fval ) ); \
-      } \
-      break; \
-    }
-
 static AV *
 sl2pl1D( SLang_Array_Type *at ) {
   AV *parray;
@@ -429,20 +415,25 @@ sl2pl1D( SLang_Array_Type *at ) {
 
   case SLANG_COMPLEX_TYPE:
     {
-      double *dptr;
-      double real, imag;
+      SV *elem;
+      double data[2];
 
-      /* not good */
-      fixme( "hacking the internals of Complex_Type arrays [1D]." );
+      /* need to check this */
+      fixme( "assuming Complex_Type elements of 1D arrays returned as foo[2] arrays." );
 
-      dptr = (double *) at->data;
       for ( i = 0; i < at->num_elements; i++ ) {
-	real = *dptr++;
-	imag = *dptr++;
-	Printf( ("  complex array1D[%d] = %g + %g i\n", i, real, imag ) );
+
+        (void) SLang_get_array_element( at, &i, data );
+	Printf( ("  complex array1D[%d] = %g + %g i\n", i, data[0], data[1] ) );
+
+	CALL_METHOD_SCALAR_SV(
+	     sv_2mortal(newSVpv("Math::Complex",0)),
+             "make",
+	     C2PL_MARG_D( data[0] ); C2PL_MARG_D( data[1] );,
+             elem );
 
 	/* do we need to bother with memory management here? */
-	av_store( parray, i, make_perl_complex(real,imag) );
+	av_store( parray, i, elem );
       }
       break;
     } /* complex */
@@ -468,23 +459,25 @@ sl2pl1D( SLang_Array_Type *at ) {
 
   case SLANG_DATATYPE_TYPE:
     {
+      int dtype; /* datatypes stored as integers not as SLtype */
+
       Printf( ("* array of datatypes\n") );
 
-      croak("Do we require >=v1.4.7 to handle datatype arrays\n");
+      for ( i = 0; i < at->num_elements; i++ ) {
 
-      //      for ( i = 0; i < at->num_elements; i++ ) {
-      //	unsigned char dtype;
-      //	(void) SLang_get_array_element( at, &i, &dtype );
-      //	Printf( ("array datatype %d = [%d]\n", i, dtype) );
-      //
-      //	/* assume that the load_string will catch any errors from the push_datatype */
-      //	(void) SLang_push_datatype( dtype );
-      //	if ( 0 != SLang_load_string( "string();") )
-      //	  croak( "Error: unable to execute 'string()' while messing with the stack\n" );
-      //	av_store( parray, i,
-      //		  sv_bless( newRV_inc( sl2pl() ), gv_stashpv("Inline::SLang::datatype",1) )
-      //		  );
-      //      } /* for: i */
+      	(void) SLang_get_array_element( at, &i, &dtype );
+
+      	Printf( ("array datatype %d = [%d] [name=%s]\n",
+		 i, dtype, SLclass_get_datatype_name(dtype)) );
+
+	/* leaks if use newRV [==newRV_inc] */
+      	av_store( parray, i,
+      		  sv_bless(
+			   newRV_noinc( newSVpv( SLclass_get_datatype_name(dtype), 0 ) ),
+			   gv_stashpv("Inline::SLang::datatype",1) )
+      		  );
+
+      } /* for: i */
 
       break;
     } /* datatype */
@@ -509,95 +502,6 @@ sl2pl1D( SLang_Array_Type *at ) {
 
 } /* sl2pl1D */
 
-/*
- * SL2PL_ARRAY2D_ITYPE( INT, int ) 
- * will create code to handle 2D array values of integer types
- *
- *   SLANG_INT_TYPE and SLANG_UINT_TYPE
- *
- * SL2PL_ARRAY2D_FTYPE( FLOAT, float ) 
- * will create code to handle 2D array values of float types
- *
- *   SLANG_FLOAT_TYPE 
- *
- * it assumes that i, j is defined as an int
- * and the array is in at and that the perl arrays
- * are in xarray & yarray
- * also nx,ny must be the axes sizes
- * and long dims[2]
- */
-
-#define SL2PL_ARRAY2D_ITYPE(stypeu,ctype) \
-  case SLANG_##stypeu##_TYPE: \
-    { \
-      ctype ival; \
-      for ( i = 0; i < nx; i++ ) { \
-	yarray = (AV *) sv_2mortal( (SV *) newAV() ); \
-	av_extend( yarray, (I32) ny ); \
- \
-	dims[0] = i; \
-	for ( j = 0; j < ny; j++ ) { \
-	  dims[1] = j; \
- \
-	  (void) SLang_get_array_element( at, dims, &ival ); \
-	  Printf( ("  ctype array2D[%d,%d] = %i\n", i, j, ival ) ); \
-	  av_store( yarray, j, newSViv( ival ) ); \
- \
-	} /* for: j */ \
- \
-	av_store( xarray, i, newRV_inc( (SV *) yarray ) ); \
- \
-      } /* for: i */ \
-      break; \
-    } \
- \
-  case SLANG_U##stypeu##_TYPE: \
-    { \
-      unsigned ctype ival; \
-      for ( i = 0; i < nx; i++ ) { \
-	yarray = (AV *) sv_2mortal( (SV *) newAV() ); \
-	av_extend( yarray, (I32) ny ); \
- \
-	dims[0] = i; \
-	for ( j = 0; j < ny; j++ ) { \
-	  dims[1] = j; \
- \
-	  (void) SLang_get_array_element( at, dims, &ival ); \
-	  Printf( ("  unsigned ctype array2D[%d,%d] = %i\n", i, j, ival ) ); \
-	  av_store( yarray, j, newSVuv( ival ) ); \
- \
-	} /* for: j */ \
- \
-	av_store( xarray, i, newRV_inc( (SV *) yarray ) ); \
- \
-      } /* for: i */ \
-      break; \
-    } 
-
-#define SL2PL_ARRAY2D_FTYPE(stypeu,ctype) \
-  case SLANG_##stypeu##_TYPE: \
-    { \
-      ctype fval; \
-      for ( i = 0; i < nx; i++ ) { \
-	yarray = (AV *) sv_2mortal( (SV *) newAV() ); \
-	av_extend( yarray, (I32) ny ); \
- \
-	dims[0] = i; \
-	for ( j = 0; j < ny; j++ ) { \
-	  dims[1] = j; \
- \
-	  (void) SLang_get_array_element( at, dims, &fval ); \
-	  Printf( ("  ctype array2D[%d,%d] = %g\n", i, j, fval ) ); \
-	  av_store( yarray, j, newSVnv( fval ) ); \
- \
-	} /* for: j */ \
- \
-	av_store( xarray, i, newRV_inc( (SV *) yarray ) ); \
- \
-      } /* for: i */ \
-      break; \
-    }
-
 static AV *
 sl2pl2D( SLang_Array_Type *at ) { 
 
@@ -655,13 +559,12 @@ sl2pl2D( SLang_Array_Type *at ) {
 
   case SLANG_COMPLEX_TYPE:
     {
-      double *dptr;
-      double real, imag;
+      double data[2];
+      SV *elem;
 
-      /* not good */
-      fixme( "hacking the internals of Complex_Type arrays [2D]." );
+      /* need to check this */
+      fixme( "assuming Complex_Type elements of 2D arrays returned as foo[2] arrays." );
 
-      dptr = (double *) at->data;
       for ( i = 0; i < nx; i++ ) {
 	yarray = (AV *) sv_2mortal( (SV *) newAV() );
 	av_extend( yarray, (I32) ny );
@@ -670,14 +573,16 @@ sl2pl2D( SLang_Array_Type *at ) {
 	for ( j = 0; j < ny; j++ ) {
 	  dims[1] = j;
 	  
-	  /* assume we're looping the correct way */
-	  real = *dptr++;
-	  imag = *dptr++;
-
-	  Printf( ("  complex array2D[%d,%d] = %g + %g i\n", i, j, real, imag ) );
+	  (void) SLang_get_array_element( at, dims, data );
+	  Printf( ("  complex array2D[%d,%d] = %g + %g i\n", i, j, data[0], data[1] ) );
 
 	  /* do we need to bother with memory management here? */
-	  av_store( yarray, j, make_perl_complex(real,imag) );
+          CALL_METHOD_SCALAR_SV(
+	     sv_2mortal(newSVpv("Math::Complex",0)),
+             "make",
+	     C2PL_MARG_D( data[0] ); C2PL_MARG_D( data[1] );,
+             elem );
+	  av_store( yarray, j, elem );
 	} /* for: j */
 
 	av_store( xarray, i, newRV_inc( (SV *) yarray ) );
@@ -687,6 +592,40 @@ sl2pl2D( SLang_Array_Type *at ) {
       break;
     } /* COMPLEX-2D */
 
+  case SLANG_DATATYPE_TYPE:
+    {
+      int dtype; /* datatypes stored as integers not as SLtype */
+
+      Printf( ("* 2D array of datatypes\n") );
+
+      for ( i = 0; i < nx; i++ ) {
+	yarray = (AV *) sv_2mortal( (SV *) newAV() );
+	av_extend( yarray, (I32) ny );
+
+	dims[0] = i;
+	for ( j = 0; j < ny; j++ ) {
+	  dims[1] = j;
+
+	  (void) SLang_get_array_element( at, dims, &dtype );
+
+	  Printf( ("  datatype array2D[%d,%d] = %d [%s]\n",
+		   i, j, dtype, SLclass_get_datatype_name(dtype) ) );
+
+	  av_store( yarray, j,
+		    sv_bless(
+			     newRV_inc( newSVpv( SLclass_get_datatype_name(dtype), 0 ) ),
+			     gv_stashpv("Inline::SLang::datatype",1) )
+		    );
+
+	} /* for: j */
+
+	av_store( xarray, i, newRV_inc( (SV *) yarray ) );
+
+      } /* for: i */
+
+      break;
+    } /* DATATYPE-2D */
+
   default:
     croak( "Error: 2D arrays of type %d are currently not supported\n", 
 	   at->data_type );
@@ -712,34 +651,6 @@ sl2plND( SLang_Array_Type *at ) {
  * - this will probably turn out to be a bad idea
  */
 
-/*
- * SL2PL_ITYPE( INT, integer, int ) 
- * will create code to handle scalar values of integer type
- *   SLANG_INT_TYPE and SLANG_UINT_TYPE
- * the second argument is the name of the SLang_pop_xxx
- * routine, and the third the c type (it's only for int/integer
- * that the second and third args are different)
- */
-
-#define SL2PL_ITYPE(stypeu,stypel,ctype) \
-  case SLANG_##stypeu##_TYPE: \
-    { \
-      ctype ival; \
-      if ( -1 == SLang_pop_##stypel ( &ival ) ) \
-	croak( "Error: unable to read stypel value from the stack\n" ); \
-      Printf( ("  stack contains: ctype = %i\n", ival ) ); \
-      return newSViv(ival); \
-    } \
-  \
-  case SLANG_U##stypeu##_TYPE: \
-    { \
-      unsigned ctype ival; \
-      if ( -1 == SLang_pop_u##stypel ( &ival ) ) \
-	croak( "Error: unable to read stypel value from the stack\n" ); \
-      Printf( ("  stack contains: unsigned ctype = %i\n", ival ) ); \
-      return newSVuv(ival); \
-    }
-
 /*
  * convert the object on the S-Lang stack to
  * a perl object.
@@ -749,7 +660,7 @@ sl2plND( SLang_Array_Type *at ) {
  * how to handle SLang_Any_Type pointers from C.
  * Perhaps we should still read from the stack and
  * recurse by pushing objects onto the stack?
- * (note: doing this when converting associative arrays)
+ * (note: actually doing this when converting associative arrays)
  */
 
 SV *
@@ -773,6 +684,29 @@ sl2pl( void ) {
       break;
     } /* NULL */
 
+  case SLANG_ANY_TYPE:
+    {
+      SLang_Any_Type *any;
+
+      /*
+       * can the stack contain one of these?
+       * (may be needed to handle recursive parse)
+       */
+      Printf( ("  stack contains: Any_Type\n") );
+
+      if ( -1 == SLang_pop_anytype( &any ) )
+	croak( "Error: unable to pop an Any_Type object from the stack\n" );
+
+      /*** UMMM, I don't know what I'm doing ***/
+      //      printf( "debug: Any_Type %d\n", any->data_type );
+      //      printf( "debug: class name %s\n", _SLclass_get_class(any->data_type)->name );
+
+      croak( "Error: can not handle an Any_Type object\n" );
+
+      SLang_free_anytype( any );
+      break;
+    } /* NULL */
+
     /* integers */
     SL2PL_ITYPE( CHAR,  char,    char )
     SL2PL_ITYPE( SHORT, short,   short )
@@ -814,13 +748,20 @@ sl2pl( void ) {
       /*
        * store as a Math::Complex object
        */
+      SV *object;
       double real, imag;
 
       if ( -1 == SLang_pop_complex( &real, &imag ) )
 	croak( "Error: unable to read complex value from the stack\n" );
       Printf( ("  stack contains: complex %g + %g i\n", real, imag ) );
 
-      return make_perl_complex( real, imag );
+      CALL_METHOD_SCALAR_SV(
+       	 sv_2mortal(newSVpv("Math::Complex",0)),
+         "make",
+	 C2PL_MARG_D( real ); C2PL_MARG_D( imag );,
+         object );
+
+      return object;
 
     } /* COMPLEX */
 
@@ -858,10 +799,10 @@ sl2pl( void ) {
     {
       HV *harray;
       AV *parray;
+      SV *arrayref;
       SLang_Array_Type *keys = NULL;
 
       int i;
-      unsigned char type;
 
       Printf( ("  stack contains an associative array\n") );
 
@@ -874,14 +815,17 @@ sl2pl( void ) {
 		"$1=();assoc_get_values($1);assoc_get_keys($1);"
                 ) )
         croak("Error: unable to parse an associative array\n");
+      _clean_slang_vars(1);
 
       /* we leave the values array on the stack */
       if ( -1 == SLang_pop_array( &keys, 0 ) )
         croak("Error: unable to pop keys array off the stack\n");
 
-      /* more a safety check in case the assoc array interface changes (unlikely) */
-      if ( keys->data_type != SLANG_STRING_TYPE )
-	croak( "Error: keys of assoc. array not returned as strings\n" );
+      /* something to check in case of errors/changes to internals of S-Lang
+       *
+       * if ( keys->data_type != SLANG_STRING_TYPE )
+       *   croak( "Error: keys of assoc. array not returned as strings\n" );
+       */
 
       /*
        * convert the value S-Lang array to a Perl array reference.
@@ -889,9 +833,10 @@ sl2pl( void ) {
        * but call sl2pl() [which means array must be left on the stack]
        * in case the array-handling code changes.
        */
-      parray = (AV *) SvRV( sl2pl() ); /* sl2pl returns an array reference */
+      arrayref = sv_2mortal( sl2pl() );
+      parray = (AV *) SvRV( arrayref );
 
-      /* create the perl hash array */
+      /* create the perl hash array (have tried making this non-mortal) */
       harray = (HV *) sv_2mortal( (SV *) newHV() );
 
       /* loop through each element, converting the values to Perl types */
@@ -903,39 +848,113 @@ sl2pl( void ) {
         (void) SLang_get_array_element( keys, &i, &keyname );
 	Printf( ( "assoc array key = [%s]\n", keyname ) );
 
-	/* get the value (all the processing has been done by sl2pl()) */
+	/*
+         * get the value (note av_fetch returns a SV **),
+         * increase the reference count, and then store it
+         * in the perl hash array
+         */
 	value = av_fetch( parray, (I32) i, 0 );
-
-	/* store it all */
+	SvREFCNT_inc( *value );
 	hv_store( harray, keyname, strlen(keyname), *value, 0 );
 
-	SLang_free_slstring( keyname );
+	SLang_free_slstring( keyname ); // is this necessary?
       }
 
-      /* do I need to free up parray ??? */
+      /* free up memory */
       SLang_free_array( keys );
       Printf( ("freed up keys array (S-Lang)\n") );
 
-      return newRV_inc( (SV *) harray ); /* this is made mortal by calling function */
+      return newRV_inc( (SV *) harray );
       break;
 
     } /* ASSOC */
 
+  case SLANG_STRUCT_TYPE:
+    {
+      SV *object;
+      SV *fieldsref;
+      AV *fields;
+      int i, nfields;
+
+      Printf( ("  stack contains: structure\n") );
+
+      /*
+       * return an Inline::SLang::struct object
+       * - handle similarly to associative arrays, in that
+       *   we take advantage of the S-Lang stack
+       */
+      if ( -1 == SLang_load_string( "$1=();get_struct_field_names($1);" ) )
+        croak("Error: unable to get the fields of a structure\n");
+
+      /* convert the item on the stack (ie the field names) to a perl array */
+      fieldsref = sv_2mortal( sl2pl() );
+      fields = (AV *) SvRV( fieldsref );
+      nfields = 1 + av_len( fields );
+      Printf( ("Number of fields in the structure = %d\n", nfields ) );
+
+      /*
+       * create the Inline::SLang::struct object 
+       */
+      CALL_METHOD_SCALAR_SV(
+        sv_2mortal(newSVpv("Inline::SLang::struct",0)), 
+        "new", XPUSHs(fieldsref);, object );
+
+      /*
+       * push all the values onto the S-Lang stack, convert them to
+       * the corresponding Perl types, and then set the corresponding
+       * field in the Perl object. Note that _push_struct_field_values()
+       * returns items in reverse order
+       *
+       * we're relying on $1 still being defined
+       */
+      if ( -1 == SLang_load_string(
+		"$2 = _push_struct_field_values($1);"
+                ) )
+        croak("Error: unable to get the values of a structure\n");
+
+      for ( i = 0; i<nfields; i++ ) {
+	SV **name;
+
+	/* get the field name */
+	name = av_fetch( fields, (I32) i, 0 );
+	Printf( ( "struct field name %d = [%s]\n", i, SvPV_nolen(*name) ) );
+
+	/* set the field: the value is read from the S-Lang stack */
+        CALL_METHOD_VOID( object,
+                          "set",
+                          XPUSHs(*name); XPUSHs( sv_2mortal(sl2pl()) ); );
+
+      } /* for: i */
+
+      _clean_slang_vars(2);
+      return object;
+      break;
+
+    } /* STRUCT */
+
   case SLANG_DATATYPE_TYPE:
     {
       /*
        * store the datatype value as a string of the name,
        * into an Inline::SLang::datatype object
        * we let S-Lang do the conversion to a string
+       *
+       * ideally would pop the datatype off the stack and then
+       * call SLclass_get_datatype_name() on it, but
+       * the pop datatype routine is only present in >= v1.4.5
+       * and can't be bothered to have 2 code pathways depending
+       * on the library version
        */
       Printf( ("  stack contains: a S-Lang datatype object\n") );
 
       /* convert the object on the stack to a string */
       if ( 0 != SLang_load_string( "string();") )
 	croak( "Error: unable to execute 'string()' while messing with the stack\n" );
+
+      /* if use newRV [== newRV_inc] then this leaks memory */
       return
 	sv_bless(
-		 newRV_inc( sl2pl() ),
+		 newRV_noinc( sl2pl() ),
 		 gv_stashpv("Inline::SLang::datatype",1)
 		 );
       break;
